(************************************************************************
 Author:	Robert Rossmair
		Wagelwitz, Rodaer Str. 6
		04688 Mutzschen
		Germany

	e-mail:	Robert.Rossmair@t-online.de
	http://home.t-online.de/home/Robert.Rossmair/

 Module:	RRScroll

 Version:	1.1.0

 History:
   01-Aug-1997: Version 1.0.0
   29-Jun-1998: Version 1.0.1
   28-Feb-1999: Version 1.1.0
   	(adds CN_CTLCOLORSCROLLBAR and WM_ERASEBKGND message handlers)

 Compatibility:	Delphi 2 & 3

 Description:   implements TrrScrollBar, a replacement for
		TScrollBar with Cardinal instead of TScrollBarInc
		(1..32767) as type for Small/LargeChange properties.

 Copyright  1997, 1998  Robert Rossmair

 Use, modification and distribution is allowed without limitation,
 warranty, or liability of any kind.
 ************************************************************************)
unit RRScroll;

interface

uses
  Windows, Messages, Consts, Classes, Controls, Forms, StdCtrls;

type
  TrrScrollBar = class(TWinControl)
  private
    FScrollInfo: TScrollInfo;
    FSmallChange: Cardinal;
    FLargeChange: Cardinal;
    FOnChange: TNotifyEvent;
    FOnScroll: TScrollEvent;
    FKind: TScrollBarKind;
    procedure DoScroll(var Message: TWMScroll);
    function GetAutoDisable: Boolean;
    function GetLargeChange: Cardinal;
    procedure SetAutoDisable(Value: Boolean);
    procedure SetKind(Value: TScrollBarKind);
    procedure SetMax(Value: Integer);
    procedure SetMin(Value: Integer);
    procedure SetPageSize(Value: UINT);
    procedure SetPosition(Value: Integer);
    procedure CNHScroll(var Message: TWMHScroll); message CN_HSCROLL;
    procedure CNVScroll(var Message: TWMVScroll); message CN_VSCROLL;
    procedure CNCtlColorScrollBar(var Message: TMessage); message CN_CTLCOLORSCROLLBAR;
    procedure WMEraseBkgnd(var Message: TWMEraseBkgnd); message WM_ERASEBKGND;
  protected
    procedure CreateParams(var Params: TCreateParams); override;
    procedure CreateWnd; override;
    procedure Change; dynamic;
    procedure Scroll(ScrollCode: TScrollCode; var ScrollPos: Integer); dynamic;
  public
    constructor Create(AOwner: TComponent); override;
    procedure SetParams(APosition, AMin, AMax: Integer; APageSize: Cardinal; AnAutoDisable: Boolean);
  published
    property Align;
    property AutoDisable: Boolean read GetAutoDisable write SetAutoDisable default True;
    property Ctl3D;
    property DragCursor;
    property DragMode;
    property Enabled;
    property Kind: TScrollBarKind read FKind write SetKind default sbHorizontal;
    property LargeChange: Cardinal read FLargeChange write FLargeChange default 1;
    property Max: Integer read FScrollInfo.nMax write SetMax default 100;
    property Min: Integer read FScrollInfo.nMin write SetMin default 0;
    property PageSize: UINT read FScrollInfo.nPage write SetPageSize default 0;
    property ParentCtl3D;
    property ParentShowHint;
    property PopupMenu;
    property Position: Integer read FScrollInfo.nPos write SetPosition default 0;
    property ShowHint;
    property SmallChange: Cardinal read FSmallChange write FSmallChange default 1;
    property TabOrder;
    property TabStop default True;
    property Visible;
    property OnChange: TNotifyEvent read FOnChange write FOnChange;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyUp;
    property OnScroll: TScrollEvent read FOnScroll write FOnScroll;
    property OnStartDrag;
  end;

procedure Register;

implementation

constructor TrrScrollBar.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  Width := 121;
  Height := GetSystemMetrics(SM_CYHSCROLL);
  TabStop := True;
  ControlStyle := [csFramed, csDoubleClicks];
  FKind := sbHorizontal;
  FScrollInfo.cbSize := SizeOf(FScrollInfo);
  FScrollInfo.nMax := 100;
  FSmallChange := 1;
  FLargeChange := 1;
end;

procedure TrrScrollBar.CreateParams(var Params: TCreateParams);
const
  Kinds: array[TScrollBarKind] of LongInt = (SBS_HORZ, SBS_VERT);
begin
  inherited CreateParams(Params);
  CreateSubClass(Params, 'SCROLLBAR');
  with Params do Style := Style or Kinds[FKind];
end;

procedure TrrScrollBar.CreateWnd;
begin
  inherited CreateWnd;
  FScrollInfo.fMask := SIF_RANGE + SIF_PAGE + SIF_POS + SIF_DISABLENOSCROLL;
  SetScrollInfo(Handle, SB_CTL, FScrollInfo, True);
end;

procedure TrrScrollBar.SetKind(Value: TScrollBarKind);
begin
  if FKind <> Value then
  begin
    FKind := Value;
    if not (csLoading in ComponentState) then SetBounds(Left, Top, Height, Width);
    RecreateWnd;
  end;
end;

function TrrScrollBar.GetAutoDisable: Boolean;
begin
  Result := (FScrollInfo.fMask and SIF_DISABLENOSCROLL) <> 0;
end;

function TrrScrollBar.GetLargeChange: Cardinal;
begin
  Result := FLargeChange;
  if Result = 0 then Result := FScrollInfo.nPage;
end;

procedure TrrScrollBar.SetParams(APosition, AMin, AMax: Integer; APageSize: Cardinal; AnAutoDisable: Boolean);
var
  Mask: UINT;
begin
  Mask := 0;
  if AMax < AMin then
{$IFDEF VER90}
    raise EInvalidOperation.CreateRes(SScrollBarRange);
{$ELSE}
    raise EInvalidOperation.Create(SScrollBarRange);
{$ENDIF}
  if APosition < AMin then APosition := AMin;
  if APosition > AMax then APosition := AMax;
  if APageSize > AMax-AMin+1 then APageSize := AMax-AMin+1;
  if (FScrollInfo.nMin <> AMin)
  or (FScrollInfo.nMax <> AMax) then Inc(Mask, SIF_RANGE);
  if FScrollInfo.nPos <> APosition then Inc(Mask, SIF_POS);
  if FScrollInfo.nPage <> APageSize then Inc(Mask, SIF_PAGE);
  if AnAutoDisable then Inc(Mask, SIF_DISABLENOSCROLL);
  if Mask <> (FScrollInfo.fMask and SIF_DISABLENOSCROLL) then
  begin
    FScrollInfo.fMask := Mask;
    FScrollInfo.nMin := AMin;
    FScrollInfo.nMax := AMax;
    FScrollInfo.nPos := APosition;
    FScrollInfo.nPage := APageSize;
    if HandleAllocated then SetScrollInfo(Handle, SB_CTL, FScrollInfo, True);
    Change;
  end;
end;

procedure TrrScrollBar.SetPosition(Value: Integer);
begin
  with FScrollInfo do
  SetParams(Value, nMin, nMax, nPage, GetAutoDisable);
end;

procedure TrrScrollBar.SetMin(Value: Integer);
begin
  with FScrollInfo do
  SetParams(nPos, Value, nMax, nPage, GetAutoDisable);
end;

procedure TrrScrollBar.SetMax(Value: Integer);
begin
  with FScrollInfo do
  SetParams(nPos, nMin, Value, nPage, GetAutoDisable);
end;

procedure TrrScrollBar.SetPageSize(Value: UINT);
begin
  with FScrollInfo do
  SetParams(nPos, nMin, nMax, Value, GetAutoDisable);
end;

procedure TrrScrollBar.SetAutoDisable(Value: Boolean);
begin
  with FScrollInfo do
  SetParams(nPos, nMin, nMax, nPage, Value);
end;

procedure TrrScrollBar.Change;
begin
  if Assigned(FOnChange) then FOnChange(Self);
end;

procedure TrrScrollBar.Scroll(ScrollCode: TScrollCode; var ScrollPos: Integer);
begin
  if Assigned(FOnScroll) then FOnScroll(Self, ScrollCode, ScrollPos);
end;

procedure TrrScrollBar.DoScroll(var Message: TWMScroll);
var
  ScrollPos: Integer;
  NewPos: Longint;
  ScrollInfo: TScrollInfo;
begin
  with Message do
  begin
    NewPos := FScrollInfo.nPos;
    case TScrollCode(ScrollCode) of
      scLineUp:
	Dec(NewPos, FSmallChange);
      scLineDown:
	Inc(NewPos, FSmallChange);
      scPageUp:
	Dec(NewPos, GetLargeChange);
      scPageDown:
	Inc(NewPos, GetLargeChange);
      scPosition,
      scTrack:
	begin
	  ScrollInfo.cbSize := SizeOf(ScrollInfo);
	  ScrollInfo.fMask := SIF_ALL;
	  if GetScrollInfo(Handle, SB_CTL, ScrollInfo) then
	  if TScrollCode(ScrollCode) = scPosition
	    then NewPos := ScrollInfo.nPos
	    else NewPos := ScrollInfo.nTrackPos;
	end;
      scTop:
	NewPos := FScrollInfo.nMin;
      scBottom:
	NewPos := FScrollInfo.nMax;
    end;
    if NewPos < FScrollInfo.nMin then NewPos := FScrollInfo.nMin;
    if NewPos > FScrollInfo.nMax then NewPos := FScrollInfo.nMax;
    ScrollPos := NewPos;
    Scroll(TScrollCode(ScrollCode), ScrollPos);
    SetPosition(ScrollPos);
  end;
end;

procedure TrrScrollBar.CNHScroll(var Message: TWMHScroll);
begin
  DoScroll(Message);
end;

procedure TrrScrollBar.CNVScroll(var Message: TWMVScroll);
begin
  DoScroll(Message);
end;

procedure TrrScrollBar.CNCtlColorScrollBar(var Message: TMessage);
begin
  with Message do
    CallWindowProc(DefWndProc, Handle, Msg, WParam, LParam);
end;

procedure TrrScrollBar.WMEraseBkgnd(var Message: TWMEraseBkgnd);
begin
  DefaultHandler(Message);
end;

const
{$IFDEF Bilch}
  PalettePage = 'Homebrew';
{$ELSE}
  PalettePage = '3rdParty';
{$ENDIF}

procedure Register;
begin
  RegisterComponents(PalettePage, [TrrScrollBar]);
end;

end.
